/*\
 *  Equation lexer for Java calculator
 *  Copyright (C) 2002  Anthony Liguori <aliguori@clam.rutgers.edu>
 *
 *  Modification: doubles replaced by BigDecimal
 *  Copyright (C) 2003  Markus Halttunen <ysxfligxrvpb@spammotel.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*/

import java.math.BigDecimal;

/** An equation lexer/parser<br>
 * Can evaluate an equation from a given string.
 * @author Anthony Liguori
 * @version 0.1.0
 */
class Equation {
  private int DIVIDE_SCALE = 16;

  /** Class constructor
   * @param data the equation to evaluate
   * @return none
   */
  public Equation(String data) {
    bits = data.toCharArray();
  }

  /** This function will evaluate an equation.  If the x variable is in the
   * equation, the function will return a value but the value of x will be
   * undefined.
   * @returns BigDecimal the value of the evaluation
   * @exception NumberFormatException this is thrown if the equation
   * is formatted improperly
   * @exception ArithmeticException this is thrown on divide-by-zero or if
   * a function is called with an invalid domain
   */
  public BigDecimal evaluate()
          throws ArithmeticException, NumberFormatException {
    pos = 0;
    BigDecimal ret = evalPlusMinus();

    if (pos != bits.length) {
      throw new NumberFormatException("Garbage at end of equation");
    }

    return ret;
  }

  /** This function will evaluate an equation using the inputted value for
   * x if an x occurs within the equation.
   * @returns BigDecimal the value of the evaluation
   * @exception NumberFormatException this is thrown if the equation
   * is formatted improperly
   * @exception ArithmeticException this is thrown on divide-by-zero or if
   * a function is called with an invalid domain
   */
  public BigDecimal evaluate(BigDecimal x)
          throws ArithmeticException, NumberFormatException {
    this.x = x;
    return evaluate();
  }

  private void skipWS() {
    while ((pos < bits.length) && Character.isWhitespace(bits[pos])) {
      pos++;
    }
  }

  /** This function will parse a real number in the input stream
   * @returns BigDecimal the value of the number
   * @exception NumberFormatException this is thrown if the number is
   * improperly formatted (such as 2.3.23)
   */
  private BigDecimal evalReal() throws NumberFormatException {
// ([0-9]*'.'[0-9]+) | ([0-9]+)
    skipWS();
    int skip = pos;

    while ((skip < bits.length) &&
            Character.isDigit(bits[skip]))
      skip++;

    if ((skip == pos) && (bits[skip] != '.')) {
      throw new NumberFormatException("skip != pos");
    }

    if ((skip < bits.length) && bits[skip] == '.') {
      int saved = ++skip;

      while ((skip < bits.length) &&
              Character.isDigit(bits[skip]))
        skip++;

      if (skip == saved) {
        throw new NumberFormatException("Invalid numeric literal");
      }
    }

    BigDecimal results = new BigDecimal(new String(bits, pos, skip - pos));

    pos = skip;

    return results;
  }

  private String peekSymbol() {
    String lhs;
    int skip = pos;

    while ((skip < bits.length) &&
            (Character.isLetterOrDigit(bits[skip]))) {
      skip++;
    }

    if (skip > pos) {
      lhs = new String(bits, pos, skip - pos);
    } else {
      lhs = new String("");
    }

    return lhs;
  }

  /** This function evaluates a symbol in the input stream
   * @returns BigDecimal the value of the symbol
   * @exception NumberFormatException this is thrown if the given symbol is
   * undefined
   */
  private BigDecimal evalSymbol()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = new BigDecimal(0.0);
    int skip = pos;

    while ((skip < bits.length) &&
            (Character.isLetterOrDigit(bits[skip]))) {
      skip++;
    }

    if (skip > pos) {
      String symbol = new String(bits, pos, skip - pos);

      if (symbol.equals("pi")) {
        lhs = new BigDecimal(Math.PI);
      } else if (symbol.equals("e")) {
        lhs = new BigDecimal(Math.E);
      } else if (symbol.equals("x")) {
        lhs = x;
      } else {
        throw new NumberFormatException("Unknown symbol: " + symbol);
      }
      pos = skip;
    }

    return lhs;
  }

  /** This function will evaluate a function call in the input stream
   * @returns BigDecimal the value of the function
   * @exception ArithmeticException this is thrown if the function returns
   * NaN - meaning the argument to the function is out of the functions
   * domain
   */
  private BigDecimal evalFunction()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = new BigDecimal(0.0);
    int skip = pos;

    while ((skip < bits.length) &&
            (Character.isLetterOrDigit(bits[skip]))) {
      skip++;
    }

    if (skip > pos) {
      String symbol = new String(bits, pos, skip - pos);

      pos = skip;

      if (symbol.equals("sin")) {
        lhs = new BigDecimal(Math.sin(evalTerm().doubleValue()));
      } else if (symbol.equals("cos")) {
        lhs = new BigDecimal(Math.cos(evalTerm().doubleValue()));
      } else if (symbol.equals("tan")) {
        lhs = new BigDecimal(Math.tan(evalTerm().doubleValue()));
      } else if (symbol.equals("asin")) {
        lhs = new BigDecimal(Math.asin(evalTerm().doubleValue()));
      } else if (symbol.equals("acos")) {
        lhs = new BigDecimal(Math.acos(evalTerm().doubleValue()));
      } else if (symbol.equals("atan")) {
        lhs = new BigDecimal(Math.atan(evalTerm().doubleValue()));
      } else if (symbol.equals("ln")) {
        lhs = new BigDecimal(Math.log(evalTerm().doubleValue()));
      }
      /*
      if (Double.isNaN(lhs)) {
        throw new ArithmeticException(symbol + ": Invalid Domain");
      }
      */
    }

    return lhs;
  }

  private boolean peekFunction() {
    String sym = peekSymbol();

    return (sym.equals("sin") ||
            sym.equals("cos") ||
            sym.equals("tan") ||
            sym.equals("asin") ||
            sym.equals("acos") ||
            sym.equals("atan") ||
            sym.equals("ln"));
  }

  /** This function evaluates a symbol, function call, or paranthetical
   * expression.
   * @returns BigDecimal the value of the term
   * @exception NumberFormatException this is thrown if paranthesis or
   * brackets are unbalanced or if a symbol or function call is malformed
   */
  private BigDecimal evalTerm()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = new BigDecimal(0.0);
    skipWS();

    switch (bits[pos]) {
      case '(':
        pos++;
        lhs = evalPlusMinus();
        skipWS();
        if ((pos < bits.length) && (bits[pos] == ')')) {
          pos++;
        } else {
          throw new NumberFormatException("Missing ')' in expression");
        }
        break;
      case '[':
        pos++;
        lhs = evalPlusMinus();
        skipWS();
        if ((pos < bits.length) && (bits[pos] == ']')) {
          pos++;
        } else {
          throw new NumberFormatException("Missing ']' in expression");
        }
        break;
      default:
        if (Character.isDigit(bits[pos]) || (bits[pos] == '.')) {
          lhs = evalReal();
        } else if (Character.isLetter(bits[pos])) {
          if (peekFunction()) {
            lhs = evalFunction();
          } else {
            lhs = evalSymbol();
          }
        } else {
          throw new NumberFormatException("Expecting Term");
        }
        break;
    }

    return lhs;
  }

  /** This function will evaluate a unary negation or no-op expression
   * @returns BigDecimal the value of the expression
   * @exception NumberFormatException this is thrown if there is no data
   * left on on the input stream
   */
  private BigDecimal evalUnary()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = new BigDecimal(0.0);
    skipWS();

    if (pos >= bits.length) {
      throw new NumberFormatException("Premature expression end");
    }

    switch (bits[pos]) {
      case '-':
        pos++;
        lhs = evalUnary().negate();
        break;
      case '+':
        pos++;
        lhs = evalUnary();
        break;
      default:
        lhs = evalTerm();
        break;
    }

    return lhs;
  }

  /** This function will evaluate an expression containing an expontential,
   * factorial, or percentage expression.
   * @returns BigDecimal the value of the expression
   * @exception ArithmeticExpression this is thrown if the factorial function
   * is given an argument out of its domain {[x] | x >= 0}
   */
  private BigDecimal evalExponent()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = evalUnary();
    boolean need_more = true;
    skipWS();

    while ((pos < bits.length) && need_more) {
      switch (bits[pos]) {
        case '^':
          pos++;
          lhs = new BigDecimal(Math.pow(lhs.doubleValue(), evalUnary().doubleValue()));
          break;
        case '!':
          int lhs_int = lhs.intValue();
          pos++;

          if (lhs.doubleValue() < 0 || lhs_int != lhs.doubleValue()) {
            throw new ArithmeticException("Factorial: Invalid domain");
          }

          lhs = new BigDecimal(1.0);
          for (int i = 2; i <= lhs_int; i++) {
            lhs = lhs.multiply(new BigDecimal(i));
          }
          break;
        case '%':
          pos++;
          lhs = lhs.divide(new BigDecimal(100), DIVIDE_SCALE, BigDecimal.ROUND_HALF_UP);
          break;
        default:
          need_more = false;
          break;
      }
    }

    return lhs;
  }

  /** This function evaluates a multiplication or division expression
   * @returns BigDecimal the value of the expression
   * @exception ArithmeticException this is thrown in the case of divide by
   * zero
   */
  private BigDecimal evalMultDivide()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = evalExponent();
    boolean need_more = true;
    skipWS();

    while ((pos < bits.length) && need_more) {
      switch (bits[pos]) {
        case '*':
          pos++;
          lhs = lhs.multiply(evalExponent());
          break;
        case '/':
          pos++;
          lhs = lhs.divide(evalExponent(), DIVIDE_SCALE, BigDecimal.ROUND_HALF_UP);  // No need to explicit check for
          // divide-by-zero, java will throw
          break;
        default:
          need_more = false;
          break;
      }
    }

    return lhs;
  }

  private BigDecimal evalPlusMinus()
          throws ArithmeticException, NumberFormatException {
    BigDecimal lhs = evalMultDivide();
    boolean need_more = true;
    skipWS();

    while ((pos < bits.length) && need_more) {
      switch (bits[pos]) {
        case '-':
          pos++;
          lhs = lhs.subtract(evalMultDivide());
          break;
        case '+':
          pos++;
          lhs = lhs.add(evalMultDivide());
          break;
        default:
          need_more = false;
          break;
      }
    }

    return lhs;
  }

  private char bits[];
  private int pos = 0;
  private BigDecimal x = new BigDecimal(0.0);
}