/*\
 *  FlexiCalc for P800
 *  Copyright (C) 2003  Markus Halttunen <ysxfligxrvpb@spammotel.com>
 *
 *  Based on Java calculator applet
 *  Copyright (C) 2002  Anthony Liguori <aliguori@clam.rutgers.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*/

import java.util.*;
import java.awt.*;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.event.*;
import java.math.BigDecimal;

/**
 * An extended calculator for the SonyEricsson P800.
 */
public class FlexiCalc extends Frame implements ActionListener {
  private static final String VERSION = "1.3";

  private static final Color COLOR_ERASE = new Color(254, 153, 163);
  private static final Color COLOR_MEMORY = new Color(136, 198, 197);
  private static final Color COLOR_DIGITS = new Color(236, 236, 236);
  private static final Color COLOR_PREFIX = new Color(215, 215, 215);
  private static final Color COLOR_FUNCTIONS = new Color(92, 101, 188);

  private static final Color COLOR_BACKGROUND = new Color(152, 184, 253);
  private static final Color COLOR_DISPLAY_BACKGROUND = new Color(210, 210, 210);

  private static final String ACTION_EQUALS = "eq";
  private static final String ACTION_EXIT = "exit";
  private static final String ACTION_COPY = "copy";
  private static final String ACTION_PASTE = "paste";
  private static final String ACTION_DELETE = "del";
  private static final String ACTION_CLEAR = "ac";
  private static final String ACTION_MEMORY_RECALL = "rcl";
  private static final String ACTION_MEMORY_STORE = "sto";
  private static final String ACTION_GET_RESULT = "ans";
  private static final String ACTION_PREFIX = "prefix";

  private static final int SEPARATOR_RIGHT = 2;
  private static final int SEPARATOR_LEFT = 5;
  private static final int SEPARATOR_TOP = 5;
  private static final int SEPARATOR_BOTTOM = 2;
  private TextArea m_textArea = new TextArea("", 5, 60, TextArea.SCROLLBARS_VERTICAL_ONLY);
  private TextField m_textField = new TextField();
  private Hashtable m_lookup = new Hashtable();
  private BigDecimal m_result = new BigDecimal(0.0);
  private TextComponent m_lastFocusedComponent = null;

  private BigDecimal m_memory = new BigDecimal(0.0);

  private static final int MAX_SCALE = 12;

  /**
   * Initialize the calculator application
   */
  public FlexiCalc() {
    setBackground(COLOR_BACKGROUND);
    m_textField.setBackground(COLOR_DISPLAY_BACKGROUND);
    // Bad-hack work around to ensure java loads _both_ classes up front
    // to avoid a nasty lag when the calculator is first used...
    Equation eq = new Equation("2 + 2");

    FocusListener focusListener = new FocusListener() {
      public void focusGained(FocusEvent event) {
        m_lastFocusedComponent = (TextComponent) event.getComponent();
      }

      public void focusLost(FocusEvent event) {
      }
    };
    m_textField.addFocusListener(focusListener);
    m_textArea.addFocusListener(focusListener);
    m_textArea.setEditable(false);
    m_textArea.setBackground(Color.white);

    setLayout(new BorderLayout());
    Panel textPanel = new Panel(new BorderLayout());

    add(BorderLayout.NORTH, textPanel);
    textPanel.add(BorderLayout.CENTER, m_textArea);
    textPanel.add(BorderLayout.SOUTH, m_textField);

    Panel buttonPanel = createButtonPanel();
    add(BorderLayout.CENTER, buttonPanel);
    setTitle("FlexiCalc " + VERSION);

    MenuBar menuBar = createMenuBar();
    setMenuBar(menuBar);
    m_textField.setFont(new Font("Dialog", Font.PLAIN, 14));
    m_textArea.setFont(new Font("Dialog", Font.PLAIN, 11));
    m_textArea.append("FlexiCalc " + VERSION + " by \nMarkus Halttunen (c) 2003\n");
    m_textField.requestFocus();

    this.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent event) {
        System.exit(0);
      }
    });
    m_textField.addKeyListener(new KeyAdapter() {
      public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_ENTER) {
          executeEquals();
        }
      }
    });
    DonationInfo.showIfNeeded("FlexiCalc", this);
  }

  /**
   * A simple helper method that creates the menubar
   * @return
   */
  private MenuBar createMenuBar() {
    MenuBar menuBar = new MenuBar();
    Menu fileMenu = new Menu("File");
    Menu editMenu = new Menu("Edit");
    Menu mathMenu = new Menu("Math");
    Menu trigMenu = new Menu("Trig");
    Menu logMenu = new Menu("Log");

    menuBar.add(fileMenu);
    menuBar.add(editMenu);
    menuBar.add(mathMenu);

    fileMenu.add(createMenuItem("Exit", ACTION_EXIT));

    editMenu.add(createMenuItem("Copy", ACTION_COPY));
    editMenu.add(createMenuItem("Paste", ACTION_PASTE));

    trigMenu.add(createMenuItem("PI", "pi"));
    trigMenu.add(createMenuItem("sin", "sin "));
    trigMenu.add(createMenuItem("cos", "cos "));
    trigMenu.add(createMenuItem("tan", "tan "));
    trigMenu.add(createMenuItem("asin", "asin "));
    trigMenu.add(createMenuItem("acos", "acos "));
    trigMenu.add(createMenuItem("atan", "atan "));

    logMenu.add(createMenuItem("ln", "ln "));
    logMenu.add(createMenuItem("e^x", "e^"));

    mathMenu.add(createMenuItem("x\262", "^2"));
    mathMenu.add(createMenuItem("x^y", "^"));
    mathMenu.add(createMenuItem("x!", "!"));
    mathMenu.add(createMenuItem("x%", "%"));
    mathMenu.addSeparator();
    mathMenu.add(trigMenu);
    mathMenu.add(logMenu);
    return menuBar;
  }

  /**
   * Method that creates a menu item and attaches the calculator as action listener
   * @param title
   * @param action
   * @return
   */
  private MenuItem createMenuItem(String title, String action) {
    MenuItem item = new MenuItem(title);
    item.addActionListener(this);
    m_lookup.put(item, action);
    return item;
  }

  /**
   * Method that creates the buttons for the calculator using the GridBagLayout
   * @return
   */
  private Panel createButtonPanel() {
    GridBagConstraints constraints = new GridBagConstraints();
    GridBagLayout layout = new GridBagLayout();
    Panel buttonPanel = new Panel(layout);
    constraints.insets.right = SEPARATOR_RIGHT;
    constraints.insets.left = 0;
    constraints.insets.bottom = SEPARATOR_BOTTOM;
    constraints.insets.top = SEPARATOR_TOP;
    constraints.fill = GridBagConstraints.BOTH;
    constraints.weightx = 1.0;
    constraints.weighty = 1.0;
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, " mr ", ACTION_MEMORY_RECALL, 0, 0, COLOR_MEMORY);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, " m+", ACTION_MEMORY_STORE, 1, 0, COLOR_MEMORY);
    addButton(buttonPanel, layout, constraints, "ans", ACTION_GET_RESULT, 2, 0, COLOR_MEMORY);
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, " \253 ", ACTION_DELETE, 3, 0, COLOR_ERASE);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, " C", ACTION_CLEAR, 4, 0, COLOR_ERASE);
    constraints.insets.top = SEPARATOR_TOP;
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, "7", "7", 0, 1, COLOR_DIGITS);
    constraints.insets.left =0 ;
    addButton(buttonPanel, layout, constraints, "8", "8", 1, 1, COLOR_DIGITS);
    addButton(buttonPanel, layout, constraints, "9", "9", 2, 1, COLOR_DIGITS);
    constraints.insets.top = 0;
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, "4", "4", 0, 2, COLOR_DIGITS);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, "5", "5", 1, 2, COLOR_DIGITS);
    addButton(buttonPanel, layout, constraints, "6", "6", 2, 2, COLOR_DIGITS);
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, "1", "1", 0, 3, COLOR_DIGITS);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, "2", "2", 1, 3, COLOR_DIGITS);
    addButton(buttonPanel, layout, constraints, "3", "3", 2, 3, COLOR_DIGITS);
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, "0", "0", 0, 4, COLOR_DIGITS);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, ".", ".", 1, 4, COLOR_DIGITS);
    addButton(buttonPanel, layout, constraints, "-/+", ACTION_PREFIX, 2, 4, COLOR_PREFIX);

    constraints.insets.left = SEPARATOR_LEFT;
    constraints.insets.top = SEPARATOR_TOP;
    addButton(buttonPanel, layout, constraints, "(", "(", 3, 1, COLOR_FUNCTIONS, Color.white);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, ")", ")", 4, 1, COLOR_FUNCTIONS, Color.white);
    constraints.insets.top = 0;
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, "\327", "*", 3, 2, COLOR_FUNCTIONS, Color.white);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, "\367", "/", 4, 2, COLOR_FUNCTIONS, Color.white);
    constraints.insets.left = SEPARATOR_LEFT;
    addButton(buttonPanel, layout, constraints, "+", "+", 3, 3, COLOR_FUNCTIONS, Color.white);
    constraints.insets.left = 0;
    addButton(buttonPanel, layout, constraints, "-", "-", 4, 3, COLOR_FUNCTIONS, Color.white);
    constraints.insets.left = SEPARATOR_LEFT;
    constraints.gridwidth = 2;
    addButton(buttonPanel, layout, constraints, "=", ACTION_EQUALS, 3, 4, COLOR_FUNCTIONS, Color.white);
    return buttonPanel;
  }

  /**
   * A simple helper method for adding a button to the calculator
   */
  private void addButton(Panel buttonPanel, GridBagLayout layout, GridBagConstraints constraints, String text, String action, int gridx, int gridy, Color bgColor) {
    addButton(buttonPanel, layout,  constraints, text, action, gridx, gridy, bgColor, Color.black);
  }
  private void addButton(Panel buttonPanel, GridBagLayout layout, GridBagConstraints constraints, String text, String action, int gridx, int gridy, Color bgColor, Color fgColor) {
    Button button = new Button(text);
    button.setFont(new Font("Dialog", Font.PLAIN, 18));
    button.setBackground(bgColor);
    button.setForeground(fgColor);
    button.addActionListener(this);
    m_lookup.put(button, action);
    buttonPanel.add(button);
    constraints.gridx = gridx;
    constraints.gridy = gridy;
    layout.setConstraints(button, constraints);
  }

  /**
   * Method that handles the all the actions the application supports
    * @param event
   */
  public void actionPerformed(ActionEvent event) {
    synchronized(m_textArea) {
      String append = (String) m_lookup.get(event.getSource());
      if (append == ACTION_EQUALS) {
        executeEquals();
      } else if (append == ACTION_CLEAR) {
        m_textField.setText("");
      } else if (append == ACTION_DELETE) {
        String currentText = m_textField.getText();
        if (currentText.length() > 0) {
          if (!"".equals(m_textField.getSelectedText())) {
            insertText("");
          } else {
            int caretPosition = m_textField.getCaretPosition();
            if (caretPosition > 0) {
              m_textField.setText(currentText.substring(0, caretPosition-1) + currentText.substring(caretPosition));
              m_textField.setCaretPosition(caretPosition-1);
            }
          }
        }
      } else if (append == ACTION_MEMORY_STORE) {
        Equation equation = new Equation(m_textField.getText());
        try {
          m_memory = equation.evaluate();
        } catch (NumberFormatException e) {
          m_textArea.append("\n" + e.getMessage());
        } catch (ArithmeticException e) {
          m_textArea.append("\n" + e.getMessage());
        }
      } else if (append == ACTION_MEMORY_RECALL) {
        insertText(handleScientificNotation(m_memory));
      } else if (append == ACTION_GET_RESULT) {
        insertText(handleScientificNotation(m_result));
      } else if (append == ACTION_COPY) {
        if (m_lastFocusedComponent != null) {
          Clipboard clipboard = getToolkit().getSystemClipboard();
          StringSelection data = new StringSelection(m_lastFocusedComponent.getSelectedText());
          clipboard.setContents(data, data);
        }
      } else if (append == ACTION_PASTE) {
        Clipboard clipboard = getToolkit().getSystemClipboard();
        Transferable clipData = clipboard.getContents(this);
        String pastedText = null;
        try {
          pastedText = (String)(clipData.getTransferData(DataFlavor.stringFlavor));
        }
        catch (Exception e) {
          pastedText = e.toString();
        }
        insertText(pastedText);
      } else if (append == ACTION_PREFIX) {
        int caretPosition = m_textField.getCaretPosition();
        String value = m_textField.getText();
        if (value.length() > 0 && value.charAt(0) == '-') {
          m_textField.setText(value.substring(1));
          m_textField.setCaretPosition(caretPosition-1);
        } else {
          m_textField.setText("-" + value);
          m_textField.setCaretPosition(caretPosition+1);
        }
      } else if (append == ACTION_EXIT) {
        System.exit(0);
      } else {
        if (isPreviousResultNeeded(append)) {
          insertText(handleScientificNotation(m_result));
        }
        insertText(append);
      }
    }
  }

  /**
   * Executes the "equals" action
   */
  private void executeEquals() {
    String text = m_textField.getText();
    Equation equation = new Equation(text);
    try {
      m_result = equation.evaluate();
      m_textArea.append("\n" + text + "  =  " + toScaledString(m_result));
      m_textField.setText("");
    } catch (NumberFormatException e) {
      m_textArea.append("\n" + e.getMessage());
    } catch (ArithmeticException e) {
      m_textArea.append("\n" + e.getMessage());
    }
  }

  /**
   * Checks if the previous result is needed for this operation
   * @return Is the previous result needed for this operation
   */
  private boolean isPreviousResultNeeded(String append) {
    if ("".equals(m_textField.getText())) {
      if (append.equals("*") || append.equals("/") || append.equals("+") || append.equals("-") || append.startsWith("^") || append.equals("!") || append.equals("%")) {
        return true;
      }
    }
    return false;
  }

  /**
   * Method that adds text to the textfield. If there is a selection on the text field, that selection is replaced.
   * @param newText
   */
  private void insertText(String newText) {
    String currentText = m_textField.getText();
    int start = m_textField.getCaretPosition();
    int end = start;
    if (!"".equals(m_textField.getSelectedText())) {
      start = m_textField.getSelectionStart();
      end = m_textField.getSelectionEnd();
    }
    m_textField.setText(currentText.substring(0, start) + newText + currentText.substring(end));
    m_textField.setCaretPosition(start + newText.length());
  }

  /**
   * Initializes the application
   * @param args
   */
  public static void main(String[] args) {
    FlexiCalc calculator = new FlexiCalc();
    calculator.setSize(240, 320);
    calculator.setVisible(true);
  }

  /**
   * A simple hack to handle the scientific notation as it is not supported by the parser as it is
   */
  private String handleScientificNotation(BigDecimal d) {
    String result = toScaledString(d);
    int expIndex = result.indexOf("E");
    if (expIndex == -1) {
      return result;
    }
    return "(" + result.substring(0, expIndex) + "*10^" + result.substring(expIndex+1) + ")";
  }

  /**
   * A help method that scales the BigDecimal as needed and returns it as a string
   * @param bigDecimal The big decimal that may need to be scaled
   * @return String presentation of the BigDecimal after possible scaling
   */
   
  private String toScaledString(BigDecimal bigDecimal) {
    if (bigDecimal.scale() > MAX_SCALE) {
      bigDecimal = bigDecimal.setScale(MAX_SCALE, BigDecimal.ROUND_HALF_UP);      
    }
    return String.valueOf(bigDecimal.doubleValue());
  }
}
